this.name           = "NShields";
this.description    = "Main script for N-Shields NPC shield simulation and visual effects";
"use strict";

//==== SETTINGS ===========================================================================

this.NPC_shields = true;
this.Thargoid_shields = true;
this.Thargoid_energy_reduction = true;

this._drawFlashers = true;
this._drawExploders = false;
this._drawSparkEffects = true;

this.NPC_flasherColours = 1; //multicolour, type 2
this.Thargoid_flasherColours = 3; //green
this._flasherScaleFactor = 2; // Should this be larger?

//---- Not (yet?) available in config interface ----

this.$npcShieldExploder = "N-Shields-explode-star-white";
this.$thargoidShieldExploder = "N-Shields-explode-ring-green";

this._shield_charge_energybank_threshold = 0.25; // imitates the corresponding global-settings.plist property

//=====================================================================================================
// Array for other NPC shipTakingDamage-like handlers which imitate player shipTakingDamage, called for all hits but amount=0 if shields still up
this._damageHandlers = [];

//==== WorldScript Event Handlers =====================================================================

this.startUp = function()
{
	this._loadSettings(); // if set using Library OXP config interface
	
	if (worldScripts.NPC_Equipment_Damage) worldScripts.NPC_Equipment_Damage.NPC_equipmentDamagedHandlers.push(this.NPC_shield_equipmentDamaged);
	
	//TODO: find shield booster stats once not every time needed?
}
this.startUpComplete = function()
{
	if (worldScripts.Lib_Config) worldScripts.Lib_Config._registerSet(this._libSettings);
}

this.shipSpawned = function(ship)
{
	if (!this.$shipHasShields(ship)) {return;}
	// TODO: recognise npc_shields = "custom"; script_info?

	if (!ship.script.NShields_damageHandlers) // may have been set up already by other OXPs
	{
		ship.script.NShields_damageHandlers = this._damageHandlers.slice(); // need to explicitly copy, not just reference
	}
	if (ship.script.shipTakingDamage)
	{
		ship.script.NShields_damageHandlers.push(ship.script.shipTakingDamage); //store existing handler, if any
	}
	
	if (ship.isFrangible) this.$configureSubEntities(ship);
	
	if (ship.isThargoid)
	{
		if (this.$configureThargoidShields(ship)) return;
	}
	this.$configureNPCShields(ship)
}

//==== Ship Setup ============================================================================

this.$shipHasShields = function (ship)
{
	if (ship.scriptInfo && ship.scriptInfo.npc_shields == "no") return false;
	return (ship.isPiloted && !(ship.primaryRole == "escape-capsule") && !ship.isStation) // isPiloted rules out buoys, rocks, weapons and debris
}

this.$configureNPCShields = function(ship)
{
	if (!this.NPC_shields)
	{
		ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.Fake_override_shipTakingDamage;
		return;
	}
	// Standard shield data:
	ship.maxForwardShield = ship.maxAftShield = 128;
	ship.forwardShieldRechargeRate = ship.aftShieldRechargeRate = 2.0;
	
	if (ship.scriptInfo) // Possible non-standard adjustments:
	{
		if (ship.scriptInfo.max_shield_forward) ship.maxForwardShield = parseInt(ship.scriptInfo.max_shield_forward);
		if (ship.scriptInfo.max_shield_aft) ship.maxAftShield = parseInt(ship.scriptInfo.max_shield_aft);
		if (ship.scriptInfo.shield_recharge_forward) ship.forwardShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_forward);
		if (ship.scriptInfo.shield_recharge_aft) ship.aftShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_aft);
		if (ship.scriptInfo.NShields_energyReduction) ship.maxEnergy -= parseInt(ship.scriptInfo.NShields_energyReduction);
	}
	
	ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.ForeAft_override_shipTakingDamage;
	ship.script.NShields_lastRechargeTime = 0;
	ship.script.NShields_rechargeTimer = new Timer(ship.script, worldScripts.NShields.ForeAft_updateRecharge.bind(ship.script), -1, 1);
	ship.script.NShields_sqRadius = Math.pow((ship.collisionRadius * 1.2), 2);
	
	this.$applyShieldUpgrades(ship);
	ship.script.NShields_flashers = this.$getShieldColours(ship);
	ship.script.NShields_exploder = this.$getShieldExploder(ship);
	
	ship.forwardShield = ship.maxForwardShield;
	ship.aftShield = ship.maxAftShield;
}
this.$configureThargoidShields = function(ship)
{
	if (!this.Thargoid_shields)
	{
		ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.Fake_override_shipTakingDamage;
		return;
	}
	ship.maxDorsalShield = ship.maxVentralShield = 192;
	ship.dorsalShieldRechargeRate = ship.ventralShieldRechargeRate = 1.5;
	if (this.Thargoid_energyReduction) ship.maxEnergy -= 192;
	if (ship.scriptInfo)
	{
		if (ship.scriptInfo.max_upper_shield) ship.maxDorsalShield = parseInt(ship.scriptInfo.max_upper_shield);
		if (ship.scriptInfo.max_lower_shield) ship.maxVentralShield = parseInt(ship.scriptInfo.max_lower_shield);
		if (ship.scriptInfo.shield_recharge_upper) ship.dorsalShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_upper);
		if (ship.scriptInfo.shield_recharge_lower) ship.ventralShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_lower);
	}
	ship.script.shipTakingDamage = ship.script.NShields_override_shipTakingDamage = this.DorsalVentral_override_shipTakingDamage;
	ship.script.NShields_lastRechargeTime = 0;
	ship.script.NShields_rechargeTimer = new Timer(ship.script, worldScripts.NShields.DorsalVentral_updateRecharge.bind(ship.script), -1, 1);
	ship.script.NShields_sqRadius = Math.pow((ship.collisionRadius * 1.2), 2);
	// No upgrades for Thargoids as the code doesn't support them any more
	ship.script.NShields_flashers = this.$getShieldColours(ship);
	ship.script.NShields_exploder = this.$getShieldExploder(ship);
	ship.dorsalShield = ship.maxDorsalShield;
	ship.ventralShield = ship.maxVentralShield;
}

this.$configureSubEntities = function(ship)
{
	if (ship.scriptInfo && ship.scriptInfo.NShields_protect_subentities == "no") return;
	//used to allow subEntities to have their own shields, but this is no longer supported
	
	for (sE in ship.subEntities)
	{
		let subent = ship.subEntities[sE];
		if (subent.script.shipTakingDamage)
		{
			subent.script.NShields_hold_shipTakingDamage = subent.script.shipTakingDamage;
		}
		subent.script.shipTakingDamage = this._subEntity_override_shipTakingDamage;
	}
}

this.$applyShieldUpgrades = function(ship) // NOTE: may need 'labels' argument for backwards-compatibility?
{
	if (ship.equipmentStatus("EQ_SHIELD_BOOSTER") == "EQUIPMENT_OK")
	{
		this.$shieldBoosterAdjustment(ship, "EQ_SHIELD_BOOSTER")
		ship.removeEquipment("EQ_SHIELD_BOOSTER")
		ship.awardEquipment("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") // dummy equipment for shipconfig compatibility
	}
	if (ship.equipmentStatus("EQ_SHIELD_ENHANCER") == "EQUIPMENT_OK")
	{
		this.$shieldBoosterAdjustment(ship, "EQ_NAVAL_SHIELD_BOOSTER")
		ship.removeEquipment("EQ_SHIELD_ENHANCER")
		ship.awardEquipment("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER")
	}
	
	if (worldScripts["SC Core"]) this.PossiblyAwardShieldCycler(ship, 0.1);
	
	if (worldScripts.ShipConfiguration_Core)
	{
		this.$shipconfAdjustment(ship);
		//worldScripts.ShipConfiguration_Core.$updateShipSpecs(ship);
	}
}

this.$shieldBoosterAdjustment = function(ship, type, reverse)
{
	//var shield_boost = this.shieldBoosters[type].shield_boost;
	//var recharge_boost = this.shieldBoosters[type].recharge_boost;
	let eq = EquipmentInfo.infoForKey(type);
	let shield_boost = parseInt(eq.scriptInfo.oolite_shield_increase);
	let recharge_boost = parseFloat(eq.scriptInfo.oolite_shield_recharge_multiplier);
	if (!recharge_boost) recharge_boost = 1; //just in case, to avoid NaN and divide-by-zero errors
	
	if (reverse) // damaging/removing shield booster, not awarding
	{
		ship.maxForwardShield -= shield_boost;
		ship.maxAftShield -= shield_boost;
		ship.forwardShieldRechargeRate /= recharge_boost;
		ship.aftShieldRechargeRate /= recharge_boost;
		if (ship.forwardShield > ship.maxForwardShield) ship.forwardShield = ship.maxForwardShield;
		if (ship.aftShield > ship.maxAftShield) ship.aftShield = ship.maxAftShield;
	}
	else
	{
		ship.maxForwardShield += shield_boost;
		ship.maxAftShield += shield_boost;
		ship.forwardShieldRechargeRate *= recharge_boost;
		ship.aftShieldRechargeRate *= recharge_boost;
		ship.forwardShield = ship.maxForwardShield; // in case this was called after another script awarded equipment
		ship.aftShield = ship.maxAftShield;
	}
}


this.$getShieldColours = function(ship)
{
	if (ship.scriptInfo)
	{
		if (ship.scriptInfo.NShields_flasherKeys) return ship.scriptInfo.NShields_flasherKeys;
		//TODO: name not number for script_info?
		if (ship.scriptInfo.NShields_flasherPreset) return this._flasherColours[ship.scriptInfo.NShields_flasherPreset];	
	}
	if (ship.isThargoid) return this._flasherColours[this.Thargoid_flasherColours];
	return this._flasherColours[this.NPC_flasherColours];
}
this.$getShieldExploder = function(ship)
{
	if (ship.collisionRadius > 128) var sizeClass = "-XL";
	else if (ship.collisionRadius > 64) var sizeClass = "-L";
	else if (ship.collisionRadius > 32) var sizeClass = "-M";
	else if (ship.collisionRadius > 16) var sizeClass = "-S";
	else var sizeClass = "-XS";
	
	if (ship.scriptInfo && ship.scriptInfo.NShields_exploders)
	{
		return ship.scriptInfo.NShields_exploders[0]+sizeClass;
	}
	if (ship.isThargoid) return this.$thargoidShieldExploder+sizeClass;
	return this.$npcShieldExploder+sizeClass;
}

//==== Shield Simulation ==================================================================
//==== For these functions 'this' must mean a ship script =================================

//---- Standard fore/aft shields: ----

this.ForeAft_override_shipTakingDamage = function (amount, whom, type, caller) //'caller' is added when protecting frangible subEntities 
{
	if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed")
	{
		//cases where, for practical purposes, no shield handling is needed - but other OXPs' handlers may still need to run
		worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
		return;
	}
	
	worldScripts.NShields.ForeAft_updateRecharge.apply(this, [clock.absoluteSeconds - this.NShields_lastRechargeTime]);
	worldScripts.NShields.$resetRechargeTimer.call(this);
	if (caller) caller.energy += amount; //heal subEntity if called from there
	else this.ship.energy += amount;
	
	if (type == "heat damage") // will not be able to locate hit, and different handling is required anyway
	{
		worldScripts.NShields.$applyHeatDamage(this.ship, amount, ["forwardShield", "aftShield"]);
		return;
	}
	
	var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom)
	var hitLocationRelative = hitLocation.subtract(this.ship.position);
	if (hitLocationRelative.dot(this.ship.vectorForward) > 0) // forward shield
	{
		this.ship.forwardShield -= amount;
		if (this.ship.forwardShield < 0)
		{
			if (caller) caller.energy += this.ship.forwardShield;
			else this.ship.energy += this.ship.forwardShield; // add negative shield value to energy
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.forwardShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.forwardShield);
			this.ship.forwardShield = 0;
		}
		else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.forwardShield, this.ship, whom)
	}
	else
	{
		this.ship.aftShield -= amount;
		if (this.ship.aftShield < 0)
		{
			if (caller) caller.energy += this.ship.aftShield;
			else this.ship.energy += this.ship.aftShield;
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.aftShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.aftShield);
			this.ship.aftShield = 0;
		}
		else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.aftShield, this.ship, whom)
	}
}
this.ForeAft_updateRecharge = function(delta) //BUG: applying this causes both shields NaN?
{
	//based as directly as possible on Oolite's player shield recharge code
	if (delta == undefined) delta = 1;
	this.NShields_lastRechargeTime = clock.absoluteSeconds;
	
	var fwdMax = this.ship.maxForwardShield;
	var aftMax = this.ship.maxAftShield;
	var shieldRechargeFwd = this.ship.forwardShieldRechargeRate * delta;
	var shieldRechargeAft = this.ship.aftShieldRechargeRate * delta;
	
	var rechargeFwd = Math.min(shieldRechargeFwd, fwdMax - this.ship.forwardShield);
	var rechargeAft = Math.min(shieldRechargeAft, aftMax - this.ship.aftShield);
	
	var energyForShields = this.ship.energy;
	if ( (this.ship.forwardShield > fwdMax * 0.25) && (this.ship.aftShield > aftMax * 0.25) )
	{
		energyForShields = Math.max(0.0, this.ship.energy - 0.1 - (this.ship.maxEnergy * worldScripts.NShields._shield_charge_energybank_threshold));
	}
	
	if (this.ship.forwardShield < this.ship.aftShield)
	{
		rechargeFwd = Math.min(rechargeFwd, energyForShields);
		rechargeAft = Math.min(rechargeAft, energyForShields - rechargeFwd);
	}
	else
	{
		rechargeAft = Math.min(rechargeAft, energyForShields);
		rechargeFwd = Math.min(rechargeFwd, energyForShields - rechargeAft);
	}
	
	this.ship.forwardShield += rechargeFwd;
	this.ship.aftShield += rechargeAft;
	this.ship.energy -= (rechargeFwd + rechargeAft);
	// Oolite's source code does something called "OOClamp_0_max_f" on the shield and energy values here. I have not included any analogous code because I don't know what the aforementioned function actually does. Hopefully this will not break things.
	if (this.ship.forwardShield == fwdMax && this.ship.aftShield == aftMax) this.NShields_rechargeTimer.stop();
}

//---- Thargoid dorsal/ventral shields: ----

this.DorsalVentral_override_shipTakingDamage = function (amount, whom, type, caller)
{
	if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed")
	{
		worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
		return;
	}
	
	worldScripts.NShields.updateRecharge_DorsalVentral.apply(this, [clock.absoluteSeconds - this.NShields_lastRechargeTime]);
	worldScripts.NShields.$resetRechargeTimer.call(this);
	if (caller) caller.energy += amount; //heal subEntity if called from there
	else this.ship.energy += amount;
	
	if (type == "heat damage")
	{
		worldScripts.NShields.$applyHeatDamage(this.ship, amount, ["dorsalShield", "ventralShield"]);
		return;
	}
	
	var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom)
	var hitLocationRelative = hitLocation.subtract(this.ship.position);
	if (hitLocationRelative.dot(this.ship.vectorUp) > 0) // dorsal shield
	{
		this.ship.dorsalShield -= amount;
		if (this.ship.dorsalShield < 0)
		{
			if (caller) caller.energy += this.ship.dorsalShield;
			else this.ship.energy += this.ship.dorsalShield; // add negative shield value to energy
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.dorsalShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.dorsalShield);
			this.ship.dorsalShield = 0;
		}
		else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.dorsalShield, this.ship, whom)
	}
	else
	{
		this.ship.ventralShield -= amount;
		if (this.ship.ventralShield < 0)
		{
			if (caller) caller.energy += this.ship.ventralShield;
			else this.ship.energy += this.ship.ventralShield;
			worldScripts.NShields.$applyDamageHandlers(this.ship, -this.ship.ventralShield, whom, type)
			if (this.ship.isVisible) worldScripts.NShields.$drawSparks(hitLocation, this.ship, -this.ship.ventralShield);
			this.ship.ventralShield = 0;
		}
		else worldScripts.NShields.$applyDamageHandlers(this.ship, 0, whom, type);
		if (this.ship.isVisible) worldScripts.NShields.$drawShieldFX(hitLocation, this.ship.ventralShield, this.ship, whom)
	}	
}
this.DorsalVentral_updateRecharge = function(delta)
{
	//This one is a bit simpler as it doesn't have to precisely mimic player shields
	if (delta == undefined) delta = 1;
	this.NShields_lastRechargeTime = clock.absoluteSeconds;
	var topMax = this.ship.maxDorsalShield;
	var btmMax = this.ship.maxVentralShield;
	var rechargeTop = Math.min(this.ship.dorsalShieldRechargeRate * delta, topMax - this.ship.dorsalShield);
	var rechargeBtm = Math.min(this.ship.ventralShieldRechargeRate * dleta, btmMax - this.ship.ventralShield);
	if (this.ship.dorsalShield < this.ship.ventralShield)
	{
		rechargeTop = Math.min(rechargeTop, this.ship.energy);
		rechargeBtm = Math.min(rechargeBtm, this.ship.energy - rechargeTop);
	}
	else
	{
		rechargeBtm = Math.min(rechargeBtm, this.ship.energy);
		rechargeTop = Math.min(rechargeTop, this.ship.energy - rechargeBtm);
	}
	this.ship.dorsalShield += rechargeTop;
	this.ship.ventralShield -= rechargeBtm;
	this.ship.energy -= (rechargeTop + rechargeBtm);
	if (this.ship.dorsalShield == topMax && this.ship.ventralShield == btmMax) this.NShields_rechargeTimer.stop();
}

//---- Sub-entity shielding: ----

this._subEntity_override_shipTakingDamage = function(amount, whom, type)
{
	this.ship.owner.script.NShields_override_shipTakingDamage(amount, whom, type, this.ship) //not simple 'shipTakingDamage' in case another OXP re-overrode the N-Shields override
	if (this.NShields_hold_shipTakingDamage) this.NShields_hold_shipTakingDamage(amount, whom, type);	
}

//---- Fake shield effects based on energy level ----

this.Fake_override_shipTakingDamage = function(amount, whom, type, caller)
{
	var fakeShield = this.ship.energy - Math.min(64, this.ship.energy / 2);
	worldScripts.NShields.$applyDamageHandlers(this.ship, amount, whom, type)
	if (this.ship.isVisible)
	{
		var hitLocation = worldScripts.NShields.$locateHit(this.ship, whom);
		if (fakeShield > 0) worldScripts.NShields.$drawShieldFX(hitLocation, fakeShield, this.ship, whom);
		else worldScripts.NShields.$drawSparks(hitLocation, this.ship, amount);
	}
}

//---- Common damage handler components: ----

this.$locateHit = function (ship, attacker) //loosely based on customshields
{
	if (!attacker)
	{
		if (player.ship.isCloaked && ship.isVisible) attacker = player.ship;
		else if (ship.script.NShields_cloaked_attacker && ship.script.NShields_cloaked_attacker.position.distanceTo(ship.position) < 25600) attacker = ship.script.NShields_cloaked_attacker;
		else attacker = ship.script.NShields_cloaked_attacker = this.$findCloakedAttacker(ship);
		if (!attacker) //still not found anyone?
		{
			log(this.name, "WARNING: attacker is null; could not find cloaked attacker")
			return ship.position;
		}
	}
	var range = attacker.position.distanceTo(ship.position) - (ship.collisionRadius + 1);
	
	var hitLocation = attacker.position.add(attacker.vectorForward.multiply(range)); // possible forward-weapon hit
	if (ship.position.squaredDistanceTo(hitLocation) > ship.script.NShields_sqRadius) // More efficient to find & compare squares as distanceTo() would involve a square root
	{
		hitLocation = attacker.position.add(attacker.vectorForward.multiply(-range)); // next try aft weapon
		if (ship.position.squaredDistanceTo(hitLocation) > ship.script.NShields_sqRadius)
		{
			hitLocation = attacker.position.add(attacker.vectorRight.multiply(range));
			if (ship.position.squaredDistanceTo(hitLocation) > ship.script.NShields_sqRadius)
			{
				hitLocation = attacker.position.add(attacker.vectorRight.multiply(-range));
				if (ship.position.squaredDistanceTo(hitLocation) > ship.script.NShields_sqRadius)
				{
					hitLocation = Vector3D.interpolate(ship.position, attacker.position, ship.collisionRadius / attacker.position.distanceTo(ship.position)) // fallback method, directly between ships
				}
			}
		}
	}
	return hitLocation;
}

this.$findCloakedAttacker = function(ship)
{
	function $isCloakedAttacker(entity)
	{
		return (entity.isShip && !entity.isPlayer && entity.isCloaked && entity.target);
	}
	var scan = system.filteredEntities(this, $isCloakedAttacker, ship, 25600);
	for (i in scan)
	{
		if (scan[i].target == ship) return scan[i];
	}
}

this.$resetRechargeTimer = function() // called with 'this' pointing to ship script
{
	this.NShields_rechargeTimer.stop()
	this.NShields_rechargeTimer.nextTime = clock.absoluteSeconds + 1;
	this.NShields_rechargeTimer.start()
}

this.$applyHeatDamage = function(ship, amount, shields)
{
	var pass = 0; //amount passed on to energy banks
	
	for (S in shields)
	{
		var shield_amt = amount / shields.length;
		ship[shields[S]] -= shield_amt;
		if (ship[shields[S]] < 0)
		{
			pass += ship[shields[S]];
			ship[shields[S]] = 0;
		}
	}
	ship.energy -= pass;
	this.$applyDamageHandlers(ship, pass, null, "heat damage")
}

this.$applyDamageHandlers = function(ship, amount, whom, type)
{
	for (f in ship.script.NShields_damageHandlers)
	{
		ship.script.NShields_damageHandlers[f].apply(ship.script, [amount, whom, type]);
	}
}

//==== Visual Effects ===================================================================

this.$drawShieldFX = function(position, shieldLevel, ship, attacker)
{
	if (shieldLevel < 1) return; // shields effectively down, no draw
	var colour_idx = Math.min(Math.floor(shieldLevel/32), (ship.script.NShields_flashers.length - 1));
	if (this._drawFlashers)
	{
		var effectType = "N-Shields_flasher_" + ship.script.NShields_flashers[colour_idx];
		var shieldFlasher = system.addVisualEffect(effectType,  position);
		if (attacker) shieldFlasher.orientation = new Vector3D().rotationTo(attacker.position.subtract(ship.position));
		shieldFlasher.script.$parent = ship;
		shieldFlasher.scale(ship.collisionRadius * this._flasherScaleFactor); //TODO: copy more sophisticated scale code?
	}
	if (this._drawExploders)
	{
		var expl = ship.spawnOne(ship.script.NShields_exploder);
		if (expl)
		{
			expl.position = position;
			expl.explode();
		}
	}
}

this.$drawSparks = function(position, ship, amount)
{
	if (!this._drawSparkEffects) return;
	
	var chance = 1;
	if (amount < 10) chance = 0.5;
	if (Math.random() <= chance)
	{
		var sparks = ship.spawnOne("NShields-sparks");
		sparks.position = position;
		sparks.explode()
	}
}

//---- Flasher lists: ----
this._multicolour1 = [
	"red",
	"orange",
	"yellow",
	"green",
	"cyan",
	"cyan",
	"blue",
	"blue",
	"purple"
];
this._multicolour2 = [
	"red2",
	"yellow2",
	"yellowgreen2",
	"green2",
	"turquoise2",
	"turquoise2",
	"blue2",
	"blue2",
	"purple2"
];
this._multicolour3 = [ //inverted: red high, blue low
	"blue3",
	"turquoise3",
	"green3",
	"yellow3",
	"yellow3",
	"orange3",
	"red3",
	"red3",
	"deep-red"
];
// next one is aimed at Thargoids but can be used on other ships
this._green = [
	"Green01",
	"Green02",
	"Green03",
	"Green04",
	"Green05",
	"green2" // re-use of green flasher from _multicolour2
];
this._blue = [
	"Blue01",
	"Blue02",
	"Blue03",
	"Blue04",
	"Blue05",
	"blue2",
	"Blue07"
];
// TODO: pink, white?
this._flasherColours = [this._multicolour1, this._multicolour2, this._multicolour3, this._green, this._blue];


//==== Compatibility with other OXPs ====================================================

//---- Handler for NPC Equipment Damage ----

this.NPC_shield_equipmentDamaged = function(equipment)
{
	var shipconf = worldScripts.ShipConfiguration_Core
	if (shipconf)
	{
		if (shipconf._frontshields.indexOf(equipment) >= 0 || shipconf._aftshields.indexOf(equipment) >= 0)
		{
			// Damage shield boosters etc. first if available
			if (this.ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") == "EQUIPMENT_OK")
			{
				this.ship.setEquipmentStatus(equipment, "EQUIPMENT_OK")
				this.ship.setEquipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER", "EQUIPMENT_DAMAGED")
				equipment = "EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER"; // so later code will immediately deactivate the damaged booster
			}
			else if (this.ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") == "EQUIPMENT_OK")
			{
				this.ship.setEquipmentStatus(equipment, "EQUIPMENT_OK")
				this.ship.setEquipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER", "EQUIPMENT_DAMAGED")
				equipment = "EQ_NSHIELDS_NPC_SHIELD_BOOSTER";
			}
			else // actually damage the basic shield generator
			{
				if (shipconf._frontshields.indexOf(equipment) >= 0) this.ship.forwardShieldRechargeRate = -0.5;
				if (shipconf._aftshields.indexOf(equipment) >= 0) this.ship.aftShieldRechargeRate = -0.5;
			}
		}
	}
	
	if (equipment == "EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") //reverse the equipment upgrade changes
	{
		worldScripts.NShields.$shieldBoosterAdjustment(this.ship, "EQ_NAVAL_SHIELD_BOOSTER", true);
	}
	
	if (equipment == "EQ_NSHIELDS_NPC_SHIELD_BOOSTER")
	{
		worldScripts.NShields.$shieldBoosterAdjustment(this.ship, "EQ_SHIELD_BOOSTER", true);
	}
}

//---- Ship Configuration shield equipment ----

this.$shipconfAdjustment = function(ship)
{
	var shipconf_fwdShield = worldScripts.ShipConfiguration_Core.$equipmentItemInUse("frontshields", ship);
	var shipconf_aftShield = worldScripts.ShipConfiguration_Core.$equipmentItemInUse("aftshields", ship);
	
	if (shipconf_fwdShield && shipconf_fwdShield != ship.script.NShields_shipconf_fwdShield) //in case we ran this function already
	{
		if (shipconf_fwdShield == "EQ_NO_FORWARD_SHIELD") ship.maxForwardShield = 0;
		else ship.maxForwardShield += parseInt(EquipmentInfo.infoForKey(shipconf_fwdShield).scriptInfo.oolite_shield_increase);
		ship.script.NShields_shipconf_fwdShield = shipconf_fwdShield;
		ship.forwardShield = ship.maxForwardShield; // Note this only works reliably as long as we only change equip at setup not in-flight
	}
	if (shipconf_aftShield && shipconf_aftShield != ship.script.NShields_shipconf_aftShield)
	{
		if (shipconf_aftShield == "EQ_NO_AFT_SHIELD") ship.maxAftShield = 0;
		else ship.maxAftShield += parseInt(EquipmentInfo.infoForKey(shipconf_aftShield).scriptInfo.oolite_shield_increase);
		ship.script.NShields_shipconf_aftShield = shipconf_aftShield;
		ship.aftShield = ship.maxAftShield;
	}
}

//---- Shield Cycler ----
this.PossiblyAwardShieldCycler = function(ship, chance)
{
	if ((ship.autoWeapons || chance >= 1) && Math.random() <= chance) //TODO: proper allocation
	{
		log (this.name, "awarding Shield Cycler to "+ship.name)
		ship.awardEquipment("EQ_SC_SHIELD_CYCLER_INTERNAL")
		if (ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") == "EQUIPMENT_OK")
		{
			//Advanced version requires Military Shield Enhancer
			ship.awardEquipment("EQ_SC_SHIELD_CYCLER_ADVANCED")
			ship.script.NShields_SC_version = 3;
			ship.script.NShields_SC_manual_version = 0; //For now, NPCs always use automatic mode //Math.round(2 + Math.random()); // 2 or 3 = Standard or Advanced Manual Configurator 
		}
		else if (ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") == "EQUIPMENT_OK")
		{
			//Standard version, requires Shield Booster
			ship.awardEquipment("EQ_SC_SHIELD_CYCLER_STANDARD")
			ship.script.NShields_SC_version = 2;
			ship.script.NShields_SC_manual_version = 0;
		}
		else
		{
			//Basic version
			ship.awardEquipment("EQ_SC_SHIELD_CYCLER_BASIC")
			ship.script.NShields_SC_version = 1;
			ship.script.NShields_SC_manual_version = 0;
		}
		//TODO: manual configurator
		log (this.name, "version awarded: "+ship.script.NShields_SC_version+", manual configurator version "+ship.script.NShields_SC_manual_version)
		
		ship.script.NShields_damageHandlers.push(this.NPC_sc_adjust) // ?
	}
}

// shipTakingDamage handler component for NPC Shield Cyclers
this.NPC_sc_adjust = function(amount, whom, type)
{
	//check cycler is active 1st?
	var adjust_input = 
	{
		caller: "N-Shields",
		setting: 3, //equal. TODO: choose appropriate setting somewhere
		functional: 100, //TODO: damage % via NPC Equipment Damage OXP?
		fwd: this.ship.forwardShield,
		fwd_threshold: (this.ship.maxForwardShield * worldScripts["Shield Cycler"]._sc_settings.sc.threshold * 0.01 ), //TODO: should NPCs be able to adjust own threshold for some cycler versions?
		aft: this.ship.aftShield,
		aft_threshold: (this.ship.maxAftShield * worldScripts["Shield Cycler"]._sc_settings.sc.threshold * 0.01 ),
		version: this.NShields_SC_version,
		manual_version: this.NShields_SC_manual_version,
		init: false,
		max_energy: this.ship.maxEnergy
	};
	
	var adjust_output = worldScripts["SC Core"]._sc_adjust(adjust_input);
	
	this.ship.forwardShield = Math.min(adjust_output.fwd, this.ship.maxForwardShield);
	this.ship.aftShield = Math.min(adjust_output.fwd, this.ship.maxAftShield);
	this.ship.energy -= adjust_output.power;
}

//---- Library OXP config interface ----

this._libSettings = {
	Name: this.name,
	Display: "N-Shields",
	Alive: "_libSettings",
	Notify: "_saveSettings",
	Bool: {
		B0: {
			Name: "NPC_shields",
			Def: true,
			Desc: "NPC fore/aft shield simulation on/off."
		},
		B1: {
			Name: "Thargoid_shields",
			Def: true,
			Desc: "Thargoid dorsal/ventral shields on/off."
		},
		B2: {
			Name: "Thargoid_energy_reduction",
			Def: true,
			Desc: "Reduce Thargoid energy banks to balance their shields."
		},
		B3: {
			Name: "_drawFlashers",
			Def: true,
			Desc: "flasher effects on/off."
		},
		B4: {
			Name: "_drawExploders",
			Def: false,
			Desc: "star-explosion effects on/off."
		},
		B5: {
			Name: "_drawSparkEffects",
			Def: true,
			Desc: "unshielded hit spark-shower effects on/off."
		}
	},
	SInt: {
		S0: {
			Name: "NPC_flasherColours",
			Def: 1,
			Min: 0,
			Max: this._flasherColours.length,
			Desc: "Colour option for standard NPC (and player) shield flasher effects.",
			Float: false
		},
		S1: {
			Name: "Thargoid_flasherColours",
			Def: 3,
			Min: 0,
			Max: this._flasherColours.length,
			Desc: "Colour option for Thargoid shield flasher effects.",
			Float: false
		},
		S2: {
			Name: "_flasherScaleFactor",
			Def: 2,
			Min: 0.2,
			Max: 8,
			Desc: "Scale factor for shield flasher effects.",
			Float: true
		},
		Info: "Colours: 0: Multicolour, 1: Multicolour (type 2), 2: Reverse Multicolour (red high, blue low), 3: Thargoid Green, 4: Blue."
	}
}
this._saveSettings = function()
{
	var std = true;
	for (i in this._libSettings.Bool)
	{
		if (this[this._libSettings.Bool[i].Name] != this._libSettings.Bool[i].Def) std = false;
	}
	for (i in this._libSettings.SInt)
	{
		if (this[this._libSettings.SInt[i].Name] != this._libSettings.SInt[i].Def) std = false;
	}
	if (!std)
	{
		var settings = {};
		for (i in this._libSettings.Bool)
		{
			let name = this._libSettings.Bool[i].Name;
			settings[name] = this[name];
		}
		for (i in this._libSettings.SInt)
		{
			let name = this._libSettings.SInt[i].Name;
			settings[name] = this[name];
		}
		missionVariables.NShields_settings = JSON.stringify(settings);
	}
	else missionVariables.NShields_settings = null;
}
this._loadSettings = function()
{
	var save = JSON.parse(missionVariables.NShields_settings);
	if (save)
	{
		for (i in this._libSettings.Bool)
		{
			let name = this._libSettings.Bool[i].Name;
			this[name] = save[name];
		}
		for (i in this._libSettings.SInt)
		{
			let name = this._libSettings.SInt[i].Name;
			this[name] = save[name];
		}
	}
}